"""
Implements the SuperVectorizer: a preprocessor to automatically apply
transformers/encoders to different types of data, without the need to
manually categorize them beforehand, or construct complex Pipelines.
"""


from typing import Dict, List, Literal, Optional, Tuple, Union
from warnings import warn

import numpy as np
import pandas as pd
import sklearn
from pandas.core.dtypes.base import ExtensionDtype
from sklearn import __version__ as sklearn_version
from sklearn.base import TransformerMixin, clone
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder

from dirty_cat import DatetimeEncoder, GapEncoder
from dirty_cat.utils import Version


def _has_missing_values(df: Union[pd.DataFrame, pd.Series]) -> bool:
    """
    Returns True if `array` contains missing values, False otherwise.
    """
    return any(df.isnull())


def _replace_false_missing(
    df: Union[pd.DataFrame, pd.Series]
) -> Union[pd.DataFrame, pd.Series]:
    """
    Takes a DataFrame or a Series, and replaces the "false missing", that is,
    strings that designate a missing value, but do not have the corresponding
    type. We convert these strings to np.nan.
    Also replaces `None` to np.nan.
    """
    # Should not replace "missing" (the string used for imputation in
    # categorical features).
    STR_NA_VALUES = [
        "null",
        "",
        "1.#QNAN",
        "#NA",
        "nan",
        "#N/A N/A",
        "-1.#QNAN",
        "<NA>",
        "-1.#IND",
        "-nan",
        "n/a",
        "-NaN",
        "1.#IND",
        "NULL",
        "NA",
        "N/A",
        "#N/A",
        "NaN",
    ]  # taken from pandas.io.parsers (version 1.1.4)
    df = df.replace(STR_NA_VALUES + [None, "?", "..."], np.nan)
    df = df.replace(r"^\s+$", np.nan, regex=True)  # Replace whitespaces
    return df


def _replace_missing_in_cat_col(ser: pd.Series, value: str = "missing") -> pd.Series:
    """
    Takes a Series with string data,
    replaces the missing values, and returns it.
    """
    ser = _replace_false_missing(ser)
    if pd.api.types.is_categorical_dtype(ser) and (value not in ser.cat.categories):
        ser = ser.cat.add_categories([value])
    ser = ser.fillna(value=value)
    return ser


OptionalTransformer = Optional[
    Union[TransformerMixin, Literal["drop", "remainder", "passthrough"]]
]


class SuperVectorizer(ColumnTransformer):
    """
    Easily transforms a heterogeneous data table (such as a dataframe) to
    a numerical array for machine learning. For this it transforms each
    column depending on its data type.
    It provides a simplified interface for the :class:`sklearn.compose.ColumnTransformer` ;
    more documentation of attributes and functions are available in its doc.

    .. versionadded:: 0.2.0

    Parameters
    ----------

    cardinality_threshold : int, default=40
        Two lists of features will be created depending on this value: strictly
        under this value, the low cardinality categorical features, and above or
        equal, the high cardinality categorical features.
        Different transformers will be applied to these two groups,
        defined by the parameters `low_card_cat_transformer` and
        `high_card_cat_transformer` respectively.
        Note: currently, missing values are counted as a single unique value
        (so they count in the cardinality).

    low_card_cat_transformer : typing.Optional[typing.Union[sklearn.base.TransformerMixin, typing.Literal["drop", "remainder", "passthrough"]]], default=None  # noqa
        Transformer used on categorical/string features with low cardinality
        (threshold is defined by `cardinality_threshold`).
        Can either be a transformer object instance (e.g. `OneHotEncoder(drop="if_binary")`),
        a `Pipeline` containing the preprocessing steps,
        'drop' for dropping the columns,
        'remainder' for applying `remainder`,
        'passthrough' to return the unencoded columns,
        or None to use the default transformer (`OneHotEncoder()`).
        Features classified under this category are imputed based on the
        strategy defined with `impute_missing`.

    high_card_cat_transformer : typing.Optional[typing.Union[sklearn.base.TransformerMixin, typing.Literal["drop", "remainder", "passthrough"]]], default=None  # noqa
        Transformer used on categorical/string features with high cardinality
        (threshold is defined by `cardinality_threshold`).
        Can either be a transformer object instance (e.g. `GapEncoder()`),
        a `Pipeline` containing the preprocessing steps,
        'drop' for dropping the columns,
        'remainder' for applying `remainder`,
        'passthrough' to return the unencoded columns,
        or None to use the default transformer (`GapEncoder(n_components=30)`).
        Features classified under this category are imputed based on the
        strategy defined with `impute_missing`.

    numerical_transformer : typing.Optional[typing.Union[sklearn.base.TransformerMixin, typing.Literal["drop", "remainder", "passthrough"]]], default=None  # noqa
        Transformer used on numerical features.
        Can either be a transformer object instance (e.g. `StandardScaler()`),
        a `Pipeline` containing the preprocessing steps,
        'drop' for dropping the columns,
        'remainder' for applying `remainder`,
        'passthrough' to return the unencoded columns,
        or None to use the default transformer (here nothing, so 'passthrough').
        Features classified under this category are not imputed at all
        (regardless of `impute_missing`).

    datetime_transformer : typing.Optional[typing.Union[sklearn.base.TransformerMixin, typing.Literal["drop", "remainder", "passthrough"]]], default=None
        Transformer used on datetime features.
        Can either be a transformer object instance (e.g. `DatetimeEncoder()`),
        a `Pipeline` containing the preprocessing steps,
        'drop' for dropping the columns,
        'remainder' for applying `remainder`,
        'passthrough' to return the unencoded columns,
        or None to use the default transformer (`DatetimeEncoder()`).
        Features classified under this category are not imputed at all
        (regardless of `impute_missing`).

    auto_cast : bool, default=True
        If set to `True`, will try to convert each column to the best possible
        data type (dtype).

    impute_missing : str, default='auto'
        When to impute missing values in categorical (textual) columns.
        'auto' will impute missing values if it is considered appropriate
        (we are using an encoder that does not support missing values and/or
        specific versions of pandas, numpy and scikit-learn).
        'force' will impute missing values in all categorical columns.
        'skip' will not impute at all.
        When imputed, missing values are replaced by the string 'missing'.
        As imputation logic for numerical features can be quite intricate,
        it is left to the user to manage.
        See also attribute `imputed_columns_`.

    remainder : typing.Union[typing.Literal["drop", "passthrough"], sklearn.base.TransformerMixin], default='drop'  # noqa
        By default, only the specified columns in `transformers` are
        transformed and combined in the output, and the non-specified
        columns are dropped. (default ``'drop'``).
        By specifying ``remainder='passthrough'``, all remaining columns that
        were not specified in `transformers` will be automatically passed
        through. This subset of columns is concatenated with the output of
        the transformers.
        By setting ``remainder`` to be an estimator, the remaining
        non-specified columns will use the ``remainder`` estimator. The
        estimator must support :term:`fit` and :term:`transform`.
        Note that using this feature requires that the DataFrame columns
        input at :term:`fit` and :term:`transform` have identical order.

    sparse_threshold: float, default=0.3
        If the output of the different transformers contains sparse matrices,
        these will be stacked as a sparse matrix if the overall density is
        lower than this value. Use sparse_threshold=0 to always return dense.
        When the transformed output consists of all dense data, the stacked
        result will be dense, and this keyword will be ignored.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

    transformer_weights : dict, default=None
        Multiplicative weights for features per transformer. The output of the
        transformer is multiplied by these weights. Keys are transformer names,
        values the weights.

    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed

    Attributes
    ----------

    transformers_: typing.List[typing.Tuple[str, typing.Union[str, sklearn.base.TransformerMixin], typing.List[str]]]  # noqa
        The collection of fitted transformers as tuples of
        (name, fitted_transformer, column). `fitted_transformer` can be an
        estimator, 'drop', or 'passthrough'. In case there were no columns
        selected, this will be an unfitted transformer.
        If there are remaining columns, the final element is a tuple of the
        form:
        ('remainder', transformer, remaining_columns) corresponding to the
        ``remainder`` parameter. If there are remaining columns, then
        ``len(transformers_)==len(transformers)+1``, otherwise
        ``len(transformers_)==len(transformers)``.

    columns_: pandas.Index
        The fitted array's columns. They are applied to the data passed
        to the `transform` method.

    types_: typing.Dict[str, type]
        A mapping of inferred types per column.
        Key is the column name, value is the inferred dtype.
        Exists only if `auto_cast=True`.

    imputed_columns_: typing.List[str]
        The list of columns in which we imputed the missing values.

    Notes
    -----
    The column order of the input data is not guaranteed to be the same
    as the output data (returned by `transform`).
    This is a due to the way the ColumnTransformer works.
    However, the output column order will always be the same for different
    calls to `transform` on a same fitted SuperVectorizer instance.
    For example, if input data has columns ['name', 'job', 'year], then output
    columns might be shuffled, e.g., ['job', 'year', 'name'], but every call
    to `transform` will return this order.
    """

    transformers_: List[Tuple[str, Union[str, TransformerMixin], List[str]]]
    columns_: pd.Index
    types_: Dict[str, type]
    imputed_columns_: List[str]

    # Override required parameters
    _required_parameters = []

    def __init__(
        self,
        *,
        cardinality_threshold: int = 40,
        low_card_cat_transformer: OptionalTransformer = None,
        high_card_cat_transformer: OptionalTransformer = None,
        numerical_transformer: OptionalTransformer = None,
        datetime_transformer: OptionalTransformer = None,
        auto_cast: bool = True,
        impute_missing: str = "auto",
        # The next parameters are inherited from ColumnTransformer
        remainder: Union[
            Literal["drop", "passthrough"], TransformerMixin
        ] = "passthrough",
        sparse_threshold: float = 0.3,
        n_jobs: int = None,
        transformer_weights=None,
        verbose: bool = False,
    ):
        super().__init__(transformers=[])

        self.cardinality_threshold = cardinality_threshold
        self.low_card_cat_transformer = low_card_cat_transformer
        self.high_card_cat_transformer = high_card_cat_transformer
        self.numerical_transformer = numerical_transformer
        self.datetime_transformer = datetime_transformer
        self.auto_cast = auto_cast
        self.impute_missing = impute_missing

        self.remainder = remainder
        self.sparse_threshold = sparse_threshold
        self.n_jobs = n_jobs
        self.transformer_weights = transformer_weights
        self.verbose = verbose

    def _more_tags(self):
        """
        Used internally by sklearn to ease the estimator checks.
        """
        return {"allow_nan": [True]}

    def _clone_transformers(self):
        """
        For each of the different transformers that can be passed,
        create the corresponding variable name with a trailing underscore,
        which is the value that will be used in `transformers`.
        We clone the instances to avoid altering them.
        See the clone function docstring.
        Note: typos are not detected here, they are left in and are detected
        down the line in `ColumnTransformer.fit_transform`.
        """
        if isinstance(self.low_card_cat_transformer, sklearn.base.TransformerMixin):
            self.low_card_cat_transformer_ = clone(self.low_card_cat_transformer)
        elif self.low_card_cat_transformer is None:
            self.low_card_cat_transformer_ = OneHotEncoder(drop="if_binary")
        elif self.low_card_cat_transformer == "remainder":
            self.low_card_cat_transformer_ = self.remainder
        else:
            self.low_card_cat_transformer_ = self.low_card_cat_transformer

        if isinstance(self.high_card_cat_transformer, sklearn.base.TransformerMixin):
            self.high_card_cat_transformer_ = clone(self.high_card_cat_transformer)
        elif self.high_card_cat_transformer is None:
            self.high_card_cat_transformer_ = GapEncoder(n_components=30)
        elif self.high_card_cat_transformer == "remainder":
            self.high_card_cat_transformer_ = self.remainder
        else:
            self.high_card_cat_transformer_ = self.high_card_cat_transformer

        if isinstance(self.numerical_transformer, sklearn.base.TransformerMixin):
            self.numerical_transformer_ = clone(self.numerical_transformer)
        elif self.numerical_transformer is None:
            self.numerical_transformer_ = "passthrough"
        elif self.numerical_transformer == "remainder":
            self.numerical_transformer_ = self.remainder
        else:
            self.numerical_transformer_ = self.numerical_transformer

        if isinstance(self.datetime_transformer, sklearn.base.TransformerMixin):
            self.datetime_transformer_ = clone(self.datetime_transformer)
        elif self.datetime_transformer is None:
            self.datetime_transformer_ = DatetimeEncoder()
        elif self.datetime_transformer == "remainder":
            self.datetime_transformer_ = self.remainder
        else:
            self.datetime_transformer_ = self.datetime_transformer

        # TODO: check that the provided transformers are valid

    def _auto_cast(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Used during fit: takes a dataframe and tries to convert
        its columns to their best possible data type.

        Parameters
        ----------
        X : pandas.DataFrame of shape (n_samples, n_features)
            The data to be transformed.

        Returns
        -------
        pandas.DataFrame
            The same pandas DataFrame, with its columns cast to the best
            possible data type.
        """
        # We replace in all columns regardless of their type,
        # as we might have some false missing
        # in numerical columns for instance.
        X = _replace_false_missing(X)

        # Handle missing values
        for col in X.columns:
            # Convert pandas' NaN value (pd.NA) to numpy NaN value (np.nan)
            # because the former tends to raise all kind of issues when dealing
            # with scikit-learn (as of version 0.24).
            if _has_missing_values(X[col]):
                # Some numerical dtypes like Int64 or Float64 only support
                # pd.NA, so they must be converted to np.float64 before.
                if pd.api.types.is_numeric_dtype(X[col]):
                    X.loc[:, col] = X[col].astype(np.float64)
                X[col].fillna(value=np.nan, inplace=True)

        # Convert to the best possible data type
        self.types_ = {}
        for col in X.columns:
            if not pd.api.types.is_datetime64_any_dtype(X[col]):
                # we don't want to cast datetime64
                try:
                    X.loc[:, col] = pd.to_numeric(X[col], errors="raise")
                except (ValueError, TypeError):
                    # Only try to convert to datetime
                    # if the variable isn't numeric.
                    try:
                        X.loc[:, col] = pd.to_datetime(
                            X[col], errors="raise", infer_datetime_format=True
                        )
                    except (ValueError, TypeError):
                        pass
            # Cast pandas dtypes to numpy dtypes
            # for earlier versions of sklearn. FIXME: which ?
            if issubclass(X[col].dtype.__class__, ExtensionDtype):
                try:
                    X.loc[:, col] = X[col].astype(X[col].dtype.type, errors="ignore")
                except (TypeError, ValueError):
                    pass
            self.types_.update({col: X[col].dtype})
        return X

    def _apply_cast(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Used during transform: takes a pandas dataframe,
        and applies the best data types learnt during fitting.

        Does the same thing as `_auto_cast`, but applies learnt info.
        """
        for col in X.columns:
            X.loc[:, col] = _replace_false_missing(X[col])
            if _has_missing_values(X[col]):
                if pd.api.types.is_numeric_dtype(X[col]):
                    X.loc[:, col] = X[col].astype(np.float64)
                X[col].fillna(value=np.nan, inplace=True)
        for col in self.imputed_columns_:
            X.loc[:, col] = _replace_missing_in_cat_col(X[col])
        for col, dtype in self.types_.items():
            X.loc[:, col] = X[col].astype(dtype)
        return X

    def fit_transform(self, X, y=None):
        """
        Fit all transformers, transform the data, and concatenate the results.
        In practice, it (1) converts features to their best possible types
        if `auto_cast=True`, (2) classify columns based on their data type,
        (3) replaces "false missing" (see function `_replace_false_missing`),
        and imputes categorical columns depending on `impute_missing`, and
        finally, transforms X.

        Parameters
        ----------
        X : {array-like, dataframe} of shape (n_samples, n_features)
            Input data, of which specified subsets are used to fit the
            transformers.
        y : array-like of shape (n_samples,), default=None
            Targets for supervised learning.

        Returns
        -------
        {array-like, sparse matrix} of shape (n_samples, sum_n_components)
            hstack of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers. If
            any result is a sparse matrix, everything will be converted to
            sparse matrices.
        """
        if self.impute_missing not in ("skip", "force", "auto"):
            raise ValueError(
                "Invalid value for `impute_missing`, expected any of "
                "{'auto', 'force', 'skip'}, "
                f"got {self.impute_missing!r}. "
            )

        self._clone_transformers()

        # Convert to pandas DataFrame if not already.
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        else:
            # Create a copy to avoid altering the original data.
            X = X.copy()

        self.columns_ = X.columns
        # If auto_cast is True, we'll find and apply the best possible type
        # to each column.
        # We'll keep the results in order to apply the types in `transform`.
        if self.auto_cast:
            X = self._auto_cast(X)

        # Select columns by dtype
        numeric_columns = X.select_dtypes(
            include=[
                "int",
                "float",
                np.float64,
                np.float32,
                np.float16,
                np.int64,
                np.int32,
                np.int16,
                np.uint64,
                np.uint32,
                np.uint16,
            ]
        ).columns.to_list()
        categorical_columns = X.select_dtypes(
            include=["string", "object", "category"]
        ).columns.to_list()
        datetime_columns = X.select_dtypes(
            include=["datetime", "datetimetz"]
        ).columns.to_list()

        # Classify categorical columns by cardinality
        _nunique_values = {  # Cache results
            col: X[col].nunique() for col in categorical_columns
        }
        low_card_cat_columns = [
            col
            for col in categorical_columns
            if _nunique_values[col] < self.cardinality_threshold
        ]
        high_card_cat_columns = [
            col
            for col in categorical_columns
            if _nunique_values[col] >= self.cardinality_threshold
        ]
        # Clear cache
        del _nunique_values

        # Next part: construct the transformers
        # Create the list of all the transformers.
        all_transformers: List[Tuple[str, OptionalTransformer, List[str]]] = [
            ("numeric", self.numerical_transformer, numeric_columns),
            ("datetime", self.datetime_transformer_, datetime_columns),
            ("low_card_cat", self.low_card_cat_transformer_, low_card_cat_columns),
            ("high_card_cat", self.high_card_cat_transformer_, high_card_cat_columns),
        ]
        # We will now filter this list, by keeping only the ones with:
        # - at least one column
        # - a valid encoder or string (filter out if None)
        self.transformers = []
        for trans in all_transformers:
            name, enc, cols = trans  # Unpack
            if len(cols) > 0 and enc is not None:
                self.transformers.append(trans)

        self.imputed_columns_ = []
        if self.impute_missing != "skip":
            # First, replace false missing
            # This is technically redundant with the call made in `_auto_cast`,
            # but we do it again anyway.
            X = _replace_false_missing(X)

            # Then, impute if suiting
            if _has_missing_values(X):
                if self.impute_missing == "force":
                    for col in X.columns:
                        # Only impute categorical columns
                        if col in categorical_columns:
                            X.loc[:, col] = _replace_missing_in_cat_col(X[col])
                            self.imputed_columns_.append(col)

                elif self.impute_missing == "auto":
                    for name, trans, cols in all_transformers:
                        impute: bool = False

                        if isinstance(trans, OneHotEncoder) and Version(
                            sklearn_version
                        ) < Version("0.24"):
                            impute = True

                        if impute:
                            for col in cols:
                                # Only impute categorical columns
                                if col in categorical_columns:
                                    X.loc[:, col] = _replace_missing_in_cat_col(X[col])
                                    self.imputed_columns_.append(col)

        # If there was missing values imputation, we cast the DataFrame again,
        # as pandas gives different types depending on whether a column has
        # missing values or not.
        if self.imputed_columns_ and self.auto_cast:
            X = self._auto_cast(X)

        if self.verbose:
            print(f"[SuperVectorizer] Assigned transformers: {self.transformers}")

        X_enc = super().fit_transform(X, y)

        # For the "remainder" columns, the `ColumnTransformer` `transformers_`
        # attribute contains the index instead of the column name,
        # so we convert the values to the appropriate column names
        # if there is less than 20 columns in the remainder.
        for i, (name, enc, cols) in enumerate(self.transformers_):
            if name == "remainder" and len(cols) < 20:
                # In this case, "cols" is a list of ints (the indices)
                cols: List[int]
                self.transformers_[i] = (name, enc, [self.columns_[j] for j in cols])

        return X_enc

    def transform(self, X) -> np.ndarray:
        """
        Transform X by applying fitted transformers on each column,
        and concatenate the results.

        Parameters
        ----------
        X : {array-like, dataframe} of shape (n_samples, n_features)
            The data to be transformed.

        Returns
        -------
        {array-like, sparse matrix} of shape (n_samples, sum_n_components)
            hstack of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers. If
            any result is a sparse matrix, everything will be converted to
            sparse matrices.
        """
        if X.shape[1] != len(self.columns_):
            raise ValueError(
                "Passed array does not match column count of "
                f"array seen during fit. Got {X.shape[1]} "
                f"columns, expected {len(self.columns_)}"
            )

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        else:
            # Create a copy to avoid altering the original data.
            X = X.copy()

        if (X.columns != self.columns_).all():
            X.columns = self.columns_

        if self.auto_cast:
            X = self._apply_cast(X)

        return super().transform(X)

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Returns clean feature names with format
        "<column_name>_<value>" if encoded by OneHotEncoder or alike,
        e.g. "job_title_Police officer", or "<column_name>" otherwise.

        Returns
        -------
        typing.List[str]
            Feature names.
        """
        if Version(sklearn_version) < Version("1.0"):
            ct_feature_names = super().get_feature_names()
        else:
            ct_feature_names = super().get_feature_names_out()
        all_trans_feature_names = []

        for name, trans, cols, _ in self._iter(fitted=True):
            if isinstance(trans, str):
                if trans == "drop":
                    continue
                elif trans == "passthrough":
                    if all(isinstance(col, int) for col in cols):
                        cols = [self.columns_[i] for i in cols]
                    all_trans_feature_names.extend(cols)
                continue
            if not hasattr(trans, "get_feature_names"):
                all_trans_feature_names.extend(cols)
            else:
                if Version(sklearn_version) < Version("1.0"):
                    trans_feature_names = trans.get_feature_names(cols)
                else:
                    trans_feature_names = trans.get_feature_names_out(cols)
                all_trans_feature_names.extend(trans_feature_names)

        if len(ct_feature_names) != len(all_trans_feature_names):
            warn("Could not extract clean feature names; returning defaults. ")
            return ct_feature_names

        return all_trans_feature_names

    def get_feature_names(self, input_features=None) -> List[str]:
        """
        Ensures compatibility with sklearn < 1.0.
        Use `get_feature_names_out` instead.
        """
        if Version(sklearn_version) >= "1.0":
            warn(
                "Following the changes in scikit-learn 1.0, "
                "get_feature_names is deprecated. "
                "Use get_feature_names_out instead. ",
                DeprecationWarning,
                stacklevel=2,
            )
        return self.get_feature_names_out(input_features)



"""
Tests fetching.py (datasets fetching off OpenML.org).
"""

import shutil
import warnings
from functools import wraps
from json import JSONDecodeError
from pathlib import Path
from unittest import mock
from unittest.mock import mock_open
from urllib.error import URLError

import pandas as pd
import pytest

from dirty_cat.datasets import fetching
from dirty_cat.datasets.fetching import (
    Details,
    Features,
    _download_and_write_openml_dataset,
    _export_gz_data_to_csv,
    _features_to_csv_format,
    _get_details,
    _get_features,
    _read_json_from_gz,
)
from dirty_cat.datasets.fetching import fetch_openml_dataset as _fetch_openml_dataset
from dirty_cat.datasets.utils import get_data_dir as _get_data_dir


@wraps(_fetch_openml_dataset)
def fetch_openml_dataset(*args, **kwargs):
    """
    Wrapper for the fetching function.
    Filters out specific warnings.
    """
    with warnings.catch_warnings():
        warnings.simplefilter(
            action="ignore",
            category=UserWarning,
        )
        return _fetch_openml_dataset(*args, **kwargs)


def get_test_data_dir() -> Path:
    return _get_data_dir("tests")


def test_fetch_openml_dataset():
    """
    Tests the ``fetch_openml_dataset()`` function in a real environment.
    Though, to avoid the test being too long,
    we will download a small dataset (<1000 entries).

    Reference:
    https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/datasets/tests/test_openml.py
    """

    # Hard-coded information about the test dataset.
    # Centralizes used information here.
    test_dataset = {
        "id": 50,
        "desc_start": "**Author**",
        "url": "https://www.openml.org/d/50",
        "csv_name": "tic-tac-toe.csv",
        "dataset_rows_count": 958,
        "dataset_columns_count": 10,
    }

    test_data_dir = get_test_data_dir()

    try:  # Try-finally block used to remove the test data directory at the end
        try:
            # First, we want to purposefully test ValueError exceptions.
            with pytest.raises(ValueError):
                assert fetch_openml_dataset(dataset_id=0, data_directory=test_data_dir)
                assert fetch_openml_dataset(
                    dataset_id=2**32, data_directory=test_data_dir
                )

            # Valid call
            returned_info = fetch_openml_dataset(
                dataset_id=test_dataset["id"],
                data_directory=test_data_dir,
            )

        except URLError:
            warnings.warn(
                "No internet connection or the website is down, test aborted."
            )
            # One could try to manually recreate the tree structure
            # created by ``fetch_openml()``,  and the
            # ``.gz`` files within in order to finish the test.
            pytest.skip(
                "Exception: Skipping this test because we encountered an "
                "issue probably related to an Internet connection problem. "
            )
            return

        assert returned_info["description"].startswith(test_dataset["desc_start"])
        assert returned_info["source"] == test_dataset["url"]
        assert returned_info["path"] == test_data_dir / test_dataset["csv_name"]

        assert returned_info["path"].is_file()

        dataset: pd.DataFrame = pd.read_csv(
            returned_info["path"], sep=",", quotechar="'", escapechar="\\"
        )

        assert dataset.shape == (
            test_dataset["dataset_rows_count"],
            test_dataset["dataset_columns_count"],
        )

    finally:
        shutil.rmtree(path=str(test_data_dir), ignore_errors=True)


@mock.patch("dirty_cat.datasets.fetching.Path.is_file")
@mock.patch("dirty_cat.datasets.fetching._get_features")
@mock.patch("dirty_cat.datasets.fetching._get_details")
@mock.patch("dirty_cat.datasets.fetching._export_gz_data_to_csv")
@mock.patch("dirty_cat.datasets.fetching._download_and_write_openml_dataset")
def test_fetch_openml_dataset_mocked(
    mock_download,
    mock_export,
    mock_get_details,
    mock_get_features,
    mock_pathlib_path_isfile,
):
    """
    Test function ``fetch_openml_dataset()``, but this time,
    we mock the functions to test its inner mechanisms.
    """

    mock_get_details.return_value = Details(
        name="Dataset_name", file_id="123456", description="Dummy dataset description."
    )
    mock_get_features.return_value = Features(
        names=["id", "name", "transaction_id", "owner", "recipient"]
    )

    test_data_dir = get_test_data_dir()

    # We test the function to see if it behaves correctly when files exist
    mock_pathlib_path_isfile.return_value = True

    fetch_openml_dataset(50, test_data_dir)

    mock_download.assert_not_called()
    mock_export.assert_not_called()
    mock_get_features.assert_not_called()
    mock_get_details.assert_called_once()

    # Reset mocks
    mock_get_details.reset_mock()

    # This time, the files do not exist
    mock_pathlib_path_isfile.return_value = False

    fetch_openml_dataset(50, test_data_dir)

    # Download should be called twice
    mock_download.assert_called_with(dataset_id=50, data_directory=get_test_data_dir())
    mock_export.assert_called_once()
    mock_get_features.assert_called_once()
    mock_get_details.assert_called_once()


@mock.patch("dirty_cat.datasets.fetching.fetch_openml")
def test__download_and_write_openml_dataset(mock_fetch_openml):
    """Tests function ``_download_and_write_openml_dataset()``."""

    dataset_id = 2

    test_data_dir = get_test_data_dir()
    _download_and_write_openml_dataset(dataset_id, test_data_dir)

    mock_fetch_openml.assert_called_once_with(
        data_id=dataset_id, data_home=str(test_data_dir), as_frame=True
    )


@mock.patch("dirty_cat.datasets.fetching.Path.is_file")
def test__read_json_from_gz(mock_pathlib_path_isfile):
    """Tests function ``_read_json_from_gz()``."""

    dummy_file_path = Path("file/path.gz")

    # Passing an invalid file path (does not exist).
    mock_pathlib_path_isfile.return_value = False
    with pytest.raises(FileNotFoundError):
        assert _read_json_from_gz(dummy_file_path)

    # Passing a valid file path,
    # but reading it does not return JSON-encoded data.
    mock_pathlib_path_isfile.return_value = True
    with mock.patch(
        "gzip.open", mock_open(read_data="This is not JSON-encoded data!")
    ) as _:
        with pytest.raises(JSONDecodeError):
            assert _read_json_from_gz(dummy_file_path)

    # Passing a valid file path, and reading it
    # returns valid JSON-encoded data.
    expected_return_value = {"data": "This is JSON-encoded data!"}
    with mock.patch(
        "gzip.open", mock_open(read_data='{"data": "This is JSON-encoded data!"}')
    ) as _:
        assert _read_json_from_gz(dummy_file_path) == expected_return_value


@mock.patch("dirty_cat.datasets.fetching._read_json_from_gz")
def test__get_details(mock_read_json_from_gz):
    """Tests function ``_get_details()``."""

    expected_return_value = Details(
        "Dataset_name", "123456", "Dummy dataset description."
    )

    mock_read_json_from_gz.return_value = {
        "data_set_description": {
            "data": "This is JSON data!",
            "name": "Dataset_name",
            "file_id": "123456",
            "description": "Dummy dataset description.",
            "extra": "extra_field",
        }
    }

    returned_value = _get_details(Path("/file/name.gz"))

    assert returned_value == expected_return_value


@mock.patch("dirty_cat.datasets.fetching._read_json_from_gz")
def test__get_features(mock_read_json_from_gz):
    """Tests function ``_get_features()``."""

    expected_return_value = Features(
        ["id", "name", "transaction_id", "owner", "recipient"]
    )

    mock_read_json_from_gz.return_value = {
        "data_features": {
            "feature": [
                {"index": "0", "name": "id", "range": "123"},
                {"index": "1", "name": "name", "dub": "dub"},
                {"index": "2", "name": "transaction_id", "type": "5"},
                {"index": "3", "name": "owner", "rights": {}},
                {"index": "4", "name": "recipient", "extra": "extra"},
            ]
        }
    }

    returned_value = _get_features(Path("/file/name.gz"))

    assert returned_value == expected_return_value


def test__export_gz_data_to_csv():
    """Tests function ``_export_gz_data_to_csv()``."""

    features = Features(
        [
            "top-left-square",
            "top-middle-square",
            "top-right-square",
            "middle-left-square",
            "middle-middle-square",
            "middle-right-square",
            "bottom-left-square",
            "bottom-middle-square",
            "bottom-right-square",
            "Class",
        ]
    )
    arff_data = (
        "% This is a comment\n"
        "@relation tic-tac-toe\n"
        "@attribute 'top-left-square' {b,o,x}\n"
        "@data\n"
        "x,x,x,x,o,o,x,o,o,positive\n"
        "x,x,x,x,o,o,o,x,o,positive\n"
    )

    dummy_gz = Path("/dummy/file.gz")
    dummy_csv = Path("/dummy/file.csv")

    with mock.patch(
        "pathlib.Path.open", mock_open(read_data="")
    ) as mock_pathlib_path_open:
        with mock.patch("gzip.open", mock_open(read_data=arff_data)) as mock_gzip_open:
            _export_gz_data_to_csv(dummy_gz, dummy_csv, features)
            mock_pathlib_path_open.assert_called_with(mode="w", encoding="utf8")
            mock_gzip_open.assert_called_with(dummy_gz, mode="rt", encoding="utf8")


def test__features_to_csv_format():
    """Tests function ``_features_to_csv_format()``."""

    features = Features(["id", "name", "transaction_id", "owner", "recipient"])
    expected_return_value = "id,name,transaction_id,owner,recipient"
    assert _features_to_csv_format(features) == expected_return_value


@mock.patch("dirty_cat.datasets.fetching.fetch_openml_dataset")
@mock.patch("dirty_cat.datasets.fetching.fetch_dataset_as_dataclass")
def test_import_all_datasets(
    mock_fetch_dataset_as_dataclass, mock_fetch_openml_dataset
):
    """Tests functions ``fetch_*()``."""

    mock_fetch_openml_dataset.return_value = {
        "name": "Example dataset",
        "description": "This is a dataset.",
        "source": "https://www.openml.org/",
        "target": "To_predict",
        "path": Path("/path/to/file.csv"),
        "read_csv_kwargs": {"a": "b"},
    }

    expected_return_value_all = fetching.DatasetAll(
        name="Example dataset",
        description="This is a dataset.",
        source="https://www.openml.org/",
        path=Path("/path/to/file.csv"),
        X=pd.DataFrame([1, 2, 3, 4]),
        y=pd.Series(
            [
                5,
            ]
        ),
    )

    expected_return_value_info_only = fetching.DatasetInfoOnly(
        name="Example dataset",
        description="This is a dataset.",
        source="https://www.openml.org/",
        target="To_predict",
        path=Path("/path/to/file.csv"),
        read_csv_kwargs={"a": "b"},
    )

    for expected_return_value, load_dataframe in zip(
        [expected_return_value_all, expected_return_value_info_only],
        [True, False],
    ):
        mock_fetch_dataset_as_dataclass.return_value = expected_return_value

        returned_value = fetching.fetch_employee_salaries(
            drop_linked=False, drop_irrelevant=False
        )
        assert expected_return_value == returned_value

        mock_fetch_openml_dataset.reset_mock()

        returned_value = fetching.fetch_road_safety()
        assert expected_return_value == returned_value

        mock_fetch_openml_dataset.reset_mock()

        returned_value = fetching.fetch_medical_charge()
        assert expected_return_value == returned_value

        mock_fetch_openml_dataset.reset_mock()

        returned_value = fetching.fetch_midwest_survey()
        assert expected_return_value == returned_value

        mock_fetch_openml_dataset.reset_mock()

        returned_value = fetching.fetch_open_payments()
        assert expected_return_value == returned_value

        mock_fetch_openml_dataset.reset_mock()

        returned_value = fetching.fetch_traffic_violations()
        assert expected_return_value == returned_value

        mock_fetch_openml_dataset.reset_mock()

        returned_value = fetching.fetch_drug_directory()
        assert expected_return_value == returned_value


if __name__ == "__main__":
    print("Tests starting")
    test_fetch_openml_dataset_mocked()
    test_fetch_openml_dataset()
    print("Tests passed")



"""
Minhash encoding of string arrays.
The principle is as follows:
  1. A string is viewed as a succession of numbers (the ASCII or UTF8
     representation of its elements).
  2. The string is then decomposed into a set of n-grams, i.e.
     n-dimensional vectors of integers.
  3. A hashing function is used to assign an integer to each n-gram.
     The minimum of the hashes over all n-grams is used in the encoding.
  4. This process is repeated with N hashing functions are used to
     form N-dimensional encodings.
Maxhash encodings can be computed similarly by taking the hashes maximum
instead.
With this procedure, strings that share many n-grams have greater
probability of having same encoding values. These encodings thus capture
morphological similarities between strings.
"""

from typing import Dict, List, Literal, Tuple

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import murmurhash3_32

from .fast_hash import ngram_min_hash
from .string_distances import get_unique_ngrams
from .utils import LRUDict, check_input


class MinHashEncoder(BaseEstimator, TransformerMixin):
    """
    Encode string categorical features as a numeric array, minhash method
    applied to ngram decomposition of strings based on ngram decomposition
    of the string.

    Parameters
    ----------
    n_components : int, default=30
        The number of dimension of encoded strings. Numbers around 300 tend to
        lead to good prediction performance, but with more computational cost.
    ngram_range : typing.Tuple[int, int], default=(2, 4)
        The lower and upper boundary of the range of n-values for different
        n-grams to be extracted. All values of n such that min_n <= n <= max_n.
        will be used.
    hashing : typing.Literal["fast", "murmur"], default=fast
        Hashing function. fast is faster but
        might have some concern with its entropy.
    minmax_hash : bool, default=False
        if True, return min hash and max hash concatenated.
    handle_missing : typing.Literal["error", "zero_impute"], default=zero_impute
        Whether to raise an error or encode missing values (NaN) with
        vectors filled with zeros.

    Attributes
    ----------
    hash_dict_ : LRUDict
        Computed hashes.

    References
    ----------
    For a detailed description of the method, see
    `Encoding high-cardinality string categorical variables
    <https://hal.inria.fr/hal-02171256v4>`_ by Cerda, Varoquaux (2019).

    """

    hash_dict_: LRUDict

    _capacity: int = 2**10

    def __init__(
        self,
        n_components: int = 30,
        ngram_range: Tuple[int, int] = (2, 4),
        hashing: Literal["fast", "murmur"] = "fast",
        minmax_hash: bool = False,
        handle_missing: Literal["error", "zero_impute"] = "zero_impute",
    ):
        self.ngram_range = ngram_range
        self.n_components = n_components
        self.hashing = hashing
        self.minmax_hash = minmax_hash
        self.handle_missing = handle_missing

    def _more_tags(self) -> Dict[str, List[str]]:
        """
        Used internally by sklearn to ease the estimator checks.
        """
        return {"X_types": ["categorical"]}

    def minhash(
        self, string: str, n_components: int, ngram_range: Tuple[int, int]
    ) -> np.array:
        """
        Encode a string using murmur hashing function.

        Parameters
        ----------
        string : str
            The string to encode.
        n_components : int
            The number of dimension of encoded string.
        ngram_range : typing.Tuple[int, int]
            The lower and upper boundaries of the range of n-values for
            different n-grams to be extracted.
            All values of n such that min_n <= n <= max_n.

        Returns
        -------
        array, shape (n_components, )
            The encoded string.
        """
        min_hashes = np.ones(n_components) * np.infty
        grams = get_unique_ngrams(string, self.ngram_range)
        if len(grams) == 0:
            grams = get_unique_ngrams(" Na ", self.ngram_range)
        for gram in grams:
            hash_array = np.array(
                [
                    murmurhash3_32("".join(gram), seed=d, positive=True)
                    for d in range(n_components)
                ]
            )
            min_hashes = np.minimum(min_hashes, hash_array)
        return min_hashes / (2**32 - 1)

    def get_fast_hash(self, string: str) -> np.array:
        """
        Encode a string with fast hashing function.
        fast hashing supports both min_hash and minmax_hash encoding.

        Parameters
        ----------
        string : str
            The string to encode.

        Returns
        -------
        np.array of shape (n_components, )
            The encoded string, using specified encoding scheme.
        """
        if self.minmax_hash:
            return np.concatenate(
                [
                    ngram_min_hash(string, self.ngram_range, seed, return_minmax=True)
                    for seed in range(self.n_components // 2)
                ]
            )
        else:
            return np.array(
                [
                    ngram_min_hash(string, self.ngram_range, seed)
                    for seed in range(self.n_components)
                ]
            )

    def fit(self, X, y=None) -> "MinHashEncoder":
        """
        Fit the MinHashEncoder to X. In practice, just initializes a dictionary
        to store encodings to speed up computation.

        Parameters
        ----------
        X : array-like, shape (n_samples, ) or (n_samples, 1)
            The string data to encode.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        MinHashEncoder
            The fitted MinHashEncoder instance.
        """
        self.count = 0
        self.hash_dict_ = LRUDict(capacity=self._capacity)
        return self

    def transform(self, X) -> np.array:
        """
        Transform X using specified encoding scheme.

        Parameters
        ----------
        X : array-like, shape (n_samples, ) or (n_samples, 1)
            The string data to encode.

        Returns
        -------
        array, shape (n_samples, n_components)
            Transformed input.
        """
        X = check_input(X)
        if self.minmax_hash:
            assert (
                self.n_components % 2 == 0
            ), "n_components should be even when minmax_hash=True. "
        if self.hashing == "murmur":
            assert (
                not self.minmax_hash
            ), 'minmax_hash is not implemented with hashing="murmur". '

        # TODO: Parallelize
        is_nan_idx = False

        if self.hashing == "fast":
            X_out = np.zeros((len(X[:]), self.n_components * X.shape[1]))
            counter = self.n_components
            for k in range(X.shape[1]):
                X_in = X[:, k].reshape(-1)
                for i, x in enumerate(X_in):
                    if isinstance(x, float):  # true if x is a missing value
                        is_nan_idx = True
                    elif x not in self.hash_dict_:
                        X_out[i, k * self.n_components : counter] = self.hash_dict_[
                            x
                        ] = self.get_fast_hash(x)
                    else:
                        X_out[i, k * self.n_components : counter] = self.hash_dict_[x]
                counter += self.n_components
        elif self.hashing == "murmur":
            X_out = np.zeros((len(X[:]), self.n_components * X.shape[1]))
            counter = self.n_components
            for k in range(X.shape[1]):
                X_in = X[:, k].reshape(-1)
                for i, x in enumerate(X_in):
                    if isinstance(x, float):
                        is_nan_idx = True
                    elif x not in self.hash_dict_:
                        X_out[i, k * self.n_components : counter] = self.hash_dict_[
                            x
                        ] = self.minhash(
                            x,
                            n_components=self.n_components,
                            ngram_range=self.ngram_range,
                        )
                    else:
                        X_out[i, k * self.n_components : counter] = self.hash_dict_[x]
                counter += self.n_components
        else:
            raise ValueError(
                f"Got hashing={self.hashing}, "
                'but expected any of {"fast", "murmur"}. '
            )

        if self.handle_missing == "error":
            if is_nan_idx:
                raise ValueError(
                    "Found missing values in input data; set "
                    "handle_missing='zero_impute' "
                    "to encode with missing values. "
                )
        elif self.handle_missing == "zero_impute":
            pass
        else:
            raise ValueError(
                f"Got handle_missing={self.handle_missing}, but expected "
                'any of {"error", "zero_impute"}. '
            )
        return X_out



import numpy as np
import pytest

from dirty_cat import target_encoder


def test_target_encoder():
    lambda_ = target_encoder.lambda_
    X1 = np.array(
        ["Red", "red", "green", "blue", "green", "green", "blue", "red"]
    ).reshape(-1, 1)
    X2 = np.array(
        ["male", "male", "female", "male", "female", "female", "female", "male"]
    ).reshape(-1, 1)
    X = np.hstack([X1, X2])

    # Case 1: binary-classification and regression
    y = np.array([1, 0, 0, 1, 0, 1, 0, 0])
    n = len(y)

    Ey_ = 3 / 8
    Eyx_ = {
        "color": {"Red": 1, "red": 0, "green": 1 / 3, "blue": 0.5},
        "gender": {"male": 0.5, "female": 0.25},
    }
    count_ = {
        "color": {"Red": 1, "red": 2, "green": 3, "blue": 2},
        "gender": {"male": 4, "female": 4},
    }

    encoder = target_encoder.TargetEncoder()
    encoder.fit(X, y)
    for j in range(X.shape[1]):
        assert np.array_equal(encoder.categories_[j], np.unique(X[:, j]))
    assert Ey_ == encoder.Ey_
    assert encoder.Eyx_[0] == Eyx_["color"]
    assert encoder.Eyx_[1] == Eyx_["gender"]

    Xtest1 = np.array(
        ["Red", "red", "blue", "green", "Red", "red", "blue", "green"]
    ).reshape(-1, 1)
    Xtest2 = np.array(
        ["male", "male", "male", "male", "female", "female", "female", "female"]
    ).reshape(-1, 1)
    Xtest = np.hstack([Xtest1, Xtest2])

    Xout = encoder.transform(Xtest)

    ans_dict = {
        var: {
            cat: Eyx_[var][cat] * lambda_(count_[var][cat], n / len(count_[var]))
            + Ey_ * (1 - lambda_(count_[var][cat], n / len(count_[var])))
            for cat in Eyx_[var]
        }
        for var in Eyx_
    }
    ans = np.zeros((n, 2))
    for j, col in enumerate(["color", "gender"]):
        for i in range(n):
            ans[i, j] = ans_dict[col][Xtest[i, j]]
    assert np.array_equal(Xout, ans)

    # Case 2: multiclass-classification
    y = np.array([1, 0, 2, 1, 0, 1, 0, 0])
    n = len(y)

    encoder = target_encoder.TargetEncoder(clf_type="multiclass-clf")
    encoder.fit(X, y)

    Ey_ = {0: 4 / 8, 1: 3 / 8, 2: 1 / 8}
    Eyx_ = {
        0: {
            "color": {"Red": 0, "red": 1, "green": 1 / 3, "blue": 0.5},
            "gender": {"male": 0.5, "female": 0.5},
        },
        1: {
            "color": {"Red": 1, "red": 0, "green": 1 / 3, "blue": 0.5},
            "gender": {"male": 0.5, "female": 0.25},
        },
        2: {
            "color": {"Red": 0, "red": 0, "green": 1 / 3, "blue": 0},
            "gender": {"male": 0, "female": 0.25},
        },
    }
    assert np.array_equal(np.unique(y), encoder.classes_)
    for k in [0, 1, 2]:
        assert Ey_[k] == encoder.Ey_[k]
        assert encoder.Eyx_[k][0] == Eyx_[k]["color"]
        assert encoder.Eyx_[k][1] == Eyx_[k]["gender"]

    count_ = {
        "color": {"Red": 1, "red": 2, "green": 3, "blue": 2},
        "gender": {"male": 4, "female": 4},
    }
    assert count_["color"] == encoder.counter_[0]
    assert count_["gender"] == encoder.counter_[1]

    ans_dict = {0: {}, 1: {}, 2: {}}
    for k in [0, 1, 2]:
        ans_dict[k] = {
            var: {
                cat: Eyx_[k][var][cat] * lambda_(count_[var][cat], n / len(count_[var]))
                + Ey_[k] * (1 - lambda_(count_[var][cat], n / len(count_[var])))
                for cat in Eyx_[k][var]
            }
            for var in Eyx_[k]
        }

    ans = np.zeros((n, 2 * 3))
    Xtest1 = np.array(
        ["Red", "red", "blue", "green", "Red", "red", "blue", "green"]
    ).reshape(-1, 1)
    Xtest2 = np.array(
        ["male", "male", "male", "male", "female", "female", "female", "female"]
    ).reshape(-1, 1)
    Xtest = np.hstack([Xtest1, Xtest2])
    for k in [0, 1, 2]:
        for j, col in enumerate(["color", "gender"]):
            for i in range(n):
                ans[i, 2 * j + k] = ans_dict[k][col][Xtest[i, j]]

    Xout = encoder.transform(Xtest)
    ans = np.zeros((n, 2 * 3))
    Xtest1 = np.array(
        ["Red", "red", "blue", "green", "Red", "red", "blue", "green"]
    ).reshape(-1, 1)
    Xtest2 = np.array(
        ["male", "male", "male", "male", "female", "female", "female", "female"]
    ).reshape(-1, 1)
    Xtest = np.hstack([Xtest1, Xtest2])
    for k in [0, 1, 2]:
        for j, col in enumerate(["color", "gender"]):
            for i in range(n):
                ans[i, j * 3 + k] = ans_dict[k][col][Xtest[i, j]]
    encoder = target_encoder.TargetEncoder(clf_type="multiclass-clf")
    encoder.fit(X, y)
    Xout = encoder.transform(Xtest)
    assert np.array_equal(Xout, ans)


def _test_missing_values(input_type, missing):
    X = [
        ["Red", "male"],
        [np.nan, "male"],
        ["green", "female"],
        ["blue", "male"],
        ["green", "female"],
        ["green", "female"],
        ["blue", "female"],
        [np.nan, np.nan],
    ]

    color_cat = ["Red", "", "green", "blue"]
    gender_cat = ["male", "", "female"]

    if input_type == "numpy":
        X = np.array(X, dtype=object)
    elif input_type == "pandas":
        pd = pytest.importorskip("pandas")
        X = pd.DataFrame(X)

    # Case 1: binary-classification and regression
    y = np.array([1, 0, 0, 1, 0, 1, 0, 0])

    Ey_ = 3 / 8
    Eyx_ = {
        "color": {"Red": 1, "": 0, "green": 1 / 3, "blue": 0.5},
        "gender": {"male": 2 / 3, "female": 0.25, "": 0},
    }
    count_ = {
        "color": {"Red": 1, "": 2, "green": 3, "blue": 2},
        "gender": {"male": 3, "female": 4, "": 1},
    }

    encoder = target_encoder.TargetEncoder(handle_missing=missing)
    if missing == "error":
        with pytest.raises(ValueError, match=r"Found missing values in input"):
            encoder.fit_transform(X, y)
        return
    elif missing == "":
        encoder.fit_transform(X, y)

        assert set(encoder.categories_[0]) == set(color_cat)
        assert set(encoder.categories_[1]) == set(gender_cat)
        assert Ey_ == encoder.Ey_
        assert encoder.Eyx_[0] == Eyx_["color"]
        assert encoder.Eyx_[1] == Eyx_["gender"]
        assert dict(encoder.counter_[0]) == count_["color"]
        assert dict(encoder.counter_[1]) == count_["gender"]
    else:
        with pytest.raises(ValueError, match=r"expected any of"):
            encoder.fit_transform(X, y)
        return


def _test_missing_values_transform(input_type, missing):
    X = [
        ["Red", "male"],
        ["red", "male"],
        ["green", "female"],
        ["blue", "male"],
        ["green", "female"],
        ["green", "female"],
        ["blue", "female"],
        ["red", "male"],
    ]

    X_test = [
        ["Red", "male"],
        [np.nan, "male"],
        ["green", "female"],
        ["blue", "male"],
        ["green", "female"],
        ["green", "female"],
        ["blue", "female"],
        [np.nan, np.nan],
    ]

    if input_type == "numpy":
        X_test = np.array(X_test, dtype=object)
    elif input_type == "pandas":
        pd = pytest.importorskip("pandas")
        X_test = pd.DataFrame(X_test)

    # Case 1: binary-classification and regression
    y = np.array([1, 0, 0, 1, 0, 1, 0, 0])

    Ey_ = 3 / 8

    encoder = target_encoder.TargetEncoder(
        handle_unknown="ignore", handle_missing=missing
    )
    if missing == "error":
        encoder.fit_transform(X, y)
        with pytest.raises(ValueError, match=r"Found missing values in input"):
            encoder.transform(X_test)
        return
    elif missing == "":
        encoder.fit_transform(X, y)
        ans = encoder.transform(X_test)

        assert np.allclose(ans[1, 0], Ey_)
        assert np.allclose(ans[-1, 0], Ey_)


def test_missing_values():
    input_types = ["list", "numpy", "pandas"]
    handle_missing = ["aaa", "error", ""]
    for input_type in input_types:
        for missing in handle_missing:
            _test_missing_values(input_type, missing)
            _test_missing_values_transform(input_type, missing)



import os
from pathlib import Path


def get_data_dir(name: str = None) -> Path:
    """
    Returns the directory in which dirty_cat looks for data.

    This is typically useful for the end-user to check
    where the data is downloaded and stored.
    """
    # Note: we stick to os.path instead of pathlib.Path because
    # it's easier to test, and the functionality is the same.
    module_path = Path(os.path.dirname(__file__)).resolve()
    data_dir = module_path / "data"
    if name is not None:
        data_dir = data_dir / name
    return data_dir



"""
Similarity encoding of string arrays.
The principle is as follows:
    1. Given an input string array X = [x1, ..., xn] with k unique categories
       [c1, ..., ck] and a similarity measure sim(s1, s2) between strings,
       we define the encoded vector of xi as [sim(xi, c1), ... , sim(xi, ck)].
       Similarity encoding of X results in a matrix with shape (n, k) that
       captures morphological similarities between string entries.
    2. To avoid dealing with high-dimensional encodings when k is high, we can
       use d << k prototypes [p1, ..., pd] with which similarities will be
       computed:  xi -> [sim(xi, p1), ..., sim(xi, pd)]. These prototypes can
       be automatically sampled from the input data (most frequent categories,
       KMeans) or provided by the user.
"""

import warnings
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import sklearn
from joblib import Parallel, delayed
from numpy.random import RandomState
from scipy import sparse
from sklearn.cluster import KMeans
from sklearn.feature_extraction.text import CountVectorizer, HashingVectorizer
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils import check_random_state
from sklearn.utils.fixes import _object_dtype_isnan

from dirty_cat.utils import Version

from .string_distances import get_ngram_count, preprocess


def _ngram_similarity_one_sample_inplace(
    x_count_vector: np.array,
    vocabulary_count_matrix: np.array,
    str_x: str,
    vocabulary_ngram_counts: np.array,
    se_dict: dict,
    unq_X: np.array,
    i: int,
    ngram_range: Tuple[int, int],
) -> None:
    """
    Update inplace a dict of similarities between a string and a vocabulary

    Parameters
    ----------
    x_count_vector : np.array
        Count vector of the sample based on the ngrams of the vocabulary
    vocabulary_count_matrix : np.array
        Count vector of the vocabulary based on its ngrams
    str_x: str
        The actual sample string
    vocabulary_ngram_counts : np.array
        Number of ngrams for each unique element of the vocabulary
    se_dict : dict
        Dictionary containing the similarities for each x in unq_X
    unq_X : np.array
        The arrays of all unique samples
    i : str
        The index of x_count_vector in the csr count matrix
    ngram_range : tuple
        n-grams to use for the decomposition, where `n_min <= n <= n_max`.
    """
    nonzero_idx = x_count_vector.indices
    nonzero_vals = x_count_vector.data

    same_grams = np.asarray(
        (vocabulary_count_matrix[:, nonzero_idx].minimum(nonzero_vals)).sum(axis=1)
    )

    all_grams = (
        get_ngram_count(str_x, ngram_range) + vocabulary_ngram_counts - same_grams
    )
    similarity = np.divide(
        same_grams, all_grams, out=np.zeros_like(same_grams), where=all_grams != 0
    )
    se_dict[unq_X[i]] = similarity.reshape(-1)


def ngram_similarity(
    X,
    cats: List[str],
    ngram_range: Tuple[int, int],
    hashing_dim: int,
    dtype: type = np.float64,
) -> np.array:
    """
    Similarity encoding for dirty categorical variables:
    Given two arrays of strings, returns the similarity encoding matrix
    of size len(X) x len(cats)

    ngram_sim(s_i, s_j) =
        ||min(ci, cj)||_1 / (||ci||_1 + ||cj||_1 - ||min(ci, cj)||_1)
    """
    min_n, max_n = ngram_range
    unq_X = np.unique(X)
    cats = np.array([" %s " % cat for cat in cats])
    unq_X_ = np.array([" %s " % x for x in unq_X])
    if not hashing_dim:
        vectorizer = CountVectorizer(
            analyzer="char", ngram_range=(min_n, max_n), dtype=dtype
        )
        vectorizer.fit(np.concatenate((cats, unq_X_)))
    else:
        vectorizer = HashingVectorizer(
            analyzer="char",
            ngram_range=(min_n, max_n),
            n_features=hashing_dim,
            norm=None,
            alternate_sign=False,
            dtype=dtype,
        )
        vectorizer.fit(X)
    count_cats = vectorizer.transform(cats)
    count_X = vectorizer.transform(unq_X_)
    # We don't need the vectorizer anymore, delete it to save memory
    del vectorizer
    sum_cats = np.asarray(count_cats.sum(axis=1))
    SE_dict = {}

    for i, x in enumerate(count_X):
        _, nonzero_idx, nonzero_vals = sparse.find(x)
        same_grams = np.asarray(
            (count_cats[:, nonzero_idx].minimum(nonzero_vals)).sum(axis=1)
        )
        all_grams = x.sum() + sum_cats - same_grams
        similarity = np.divide(same_grams, all_grams)
        SE_dict[unq_X[i]] = similarity.reshape(-1)
    # We don't need the counts anymore, delete them to save memory
    del count_cats, count_X

    out = np.empty((len(X), similarity.size), dtype=dtype)
    for x, out_row in zip(X, out):
        out_row[:] = SE_dict[x]

    return np.nan_to_num(out, copy=False)


def get_prototype_frequencies(prototypes: np.array) -> np.array:
    """
    Computes the frequencies of the values contained in prototypes
    Reverse sorts the array by the frequency
    Returns a numpy array of the values without their frequencies
    """
    uniques, counts = np.unique(prototypes, return_counts=True)
    sorted_indexes = np.argsort(counts)[::-1]
    return uniques[sorted_indexes], counts[sorted_indexes]


def get_kmeans_prototypes(
    X,
    n_prototypes: int,
    hashing_dim: int = 128,
    ngram_range: Tuple[int, int] = (3, 3),
    sparse: bool = False,
    sample_weight=None,
    random_state: Optional[Union[int, RandomState]] = None,
) -> np.array:
    """
    Computes prototypes based on:
      - dimensionality reduction (via hashing n-grams)
      - k-means clustering
      - nearest neighbor
    """
    vectorizer = HashingVectorizer(
        analyzer="char",
        norm=None,
        alternate_sign=False,
        ngram_range=ngram_range,
        n_features=hashing_dim,
    )
    projected = vectorizer.transform(X)
    if not sparse:
        projected = projected.toarray()
    kmeans = KMeans(n_clusters=n_prototypes, random_state=random_state)
    kmeans.fit(projected, sample_weight=sample_weight)
    centers = kmeans.cluster_centers_
    neighbors = NearestNeighbors()
    neighbors.fit(projected)
    indexes_prototypes = np.unique(neighbors.kneighbors(centers, 1)[-1])
    if indexes_prototypes.shape[0] < n_prototypes:
        warnings.warn(
            "Final number of unique prototypes is lower than "
            + "n_prototypes (expected). "
        )
    return np.sort(X[indexes_prototypes])


class SimilarityEncoder(OneHotEncoder):
    """
    Encode string categorical features as a numeric array.

    The input to this transformer should be an array-like of
    strings.
    The method is based on calculating the morphological similarities
    between the categories.
    This encoding is an alternative to OneHotEncoder in the case of
    dirty categorical variables.

    Parameters
    ----------
    similarity : None
        Deprecated in dirty_cat 0.3, will be removed in 0.5.
        Was used to specify the type of pairwise string similarity to use.
        Since 0.3, only the ngram similarity is supported.
    ngram_range : tuple (min_n, max_n), default=(2, 4)
        The range of values for the n_gram similarity.
    categories : typing.Union[typing.Literal["auto", "k-means", "most_frequent"], typing.List[typing.List[str]]]  # noqa
        Categories (unique values) per feature:
        - 'auto' : Determine categories automatically from the training data.
        - list : ``categories[i]`` holds the categories expected in the i-th
          column. The passed categories must be sorted and should not mix
          strings and numeric values.
        - 'most_frequent' : Computes the most frequent values for every
           categorical variable
        - 'k-means' : Computes the K nearest neighbors of K-mean centroids
           in order to choose the prototype categories
        The categories used can be found in the ``categories_`` attribute.
    dtype : number type, default np.float64
        Desired dtype of output.
    handle_unknown : 'error' or 'ignore' (default)
        Whether to raise an error or ignore if an unknown categorical feature
        is present during transform (default is to ignore). When this parameter
        is set to 'ignore' and an unknown category is encountered during
        transform, the resulting encoded columns for this feature
        will be all zeros. In the inverse transform, an unknown category
        will be denoted as None.
    handle_missing : 'error' or '' (default)
        Whether to raise an error or impute with blank string '' if missing
        values (NaN) are present during fit (default is to impute).
        When this parameter is set to '', and a missing value is encountered
        during fit_transform, the resulting encoded columns for this feature
        will be all zeros. In the inverse transform, the missing category
        will be denoted as None.
    hashing_dim : int type or None.
        If None, the base vectorizer is CountVectorizer, else it's set to
        HashingVectorizer with a number of features equal to `hashing_dim`.
    n_prototypes : number of prototype we want to use.
        Useful when `most_frequent` or `k-means` is used.
        Must be a positive non-null integer.
    random_state : either an int used as a seed, a RandomState instance or None.
        Useful when `k-means` strategy is used.
    n_jobs : int, optional
        maximum number of processes used to compute similarity matrices. Used
        only if ``fast=True`` in ``SimilarityEncoder.transform``

    Attributes
    ----------
    categories_ : typing.List[np.array]
        The categories of each feature determined during fitting
        (in order corresponding with output of ``transform``).
    _infrequent_enabled : bool, default=False
        Avoid taking into account the existence of infrequent categories.

    References
    ----------

    For a detailed description of the method, see
    `Similarity encoding for learning with dirty categorical variables
    <https://hal.inria.fr/hal-01806175>`_ by Cerda, Varoquaux, Kegl. 2018
    (accepted for publication at: Machine Learning journal, Springer).

    """

    categories_: List[np.array]
    n_features_in_: int
    random_state_: Union[int, RandomState]
    drop_idx_: np.array
    vectorizers_: List[CountVectorizer]
    vocabulary_count_matrices_: List[np.array]
    vocabulary_ngram_counts_: List[List[int]]
    _infrequent_enabled: bool

    def __init__(
        self,
        similarity: str = "ngram",
        ngram_range: Tuple[int, int] = (2, 4),
        categories: Union[
            Literal["auto", "k-means", "most_frequent"], List[List[str]]
        ] = "auto",
        dtype: type = np.float64,
        handle_unknown: Literal["error", "ignore"] = "ignore",
        handle_missing: Literal["error", ""] = "",
        hashing_dim: Optional[int] = None,
        n_prototypes: Optional[int] = None,
        random_state: Optional[Union[int, RandomState]] = None,
        n_jobs: Optional[int] = None,
    ):
        super().__init__()
        self.categories = categories
        self.dtype = dtype
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing
        self.ngram_range = ngram_range
        self.hashing_dim = hashing_dim
        self.n_prototypes = n_prototypes
        self.random_state = random_state
        self.n_jobs = n_jobs

        if similarity is not None:
            warnings.warn(
                'The "similarity" argument is deprecated since dirty_cat 0.3, '
                "and will be removed in 0.5."
                "The n-gram similarity is the only one currently supported. ",
                category=UserWarning,
                stacklevel=2,
            )
        self.similarity = None

        if not isinstance(categories, list):
            assert categories in [None, "auto", "k-means", "most_frequent"]
        if categories in ["k-means", "most_frequent"] and (
            n_prototypes is None or n_prototypes == 0
        ):
            raise ValueError(
                "n_prototypes expected None or a positive non null integer. "
            )
        if categories == "auto" and n_prototypes is not None:
            warnings.warn('n_prototypes parameter ignored with category type "auto". ')

    def get_most_frequent(self, prototypes: List[str]) -> np.array:
        """
        Get the most frequent category prototypes.

        Parameters
        ----------
        prototypes : typing.List[str]
            The list of values for a category variable.

        Returns
        -------
        np.array
            The n_prototypes most frequent values for a category variable.
        """
        values, _ = get_prototype_frequencies(prototypes)
        return values[: self.n_prototypes]

    def fit(self, X, y=None) -> "SimilarityEncoder":
        """
        Fit the SimilarityEncoder to X.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to determine the categories of each feature.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        SimilarityEncoder
            The fitted SimilarityEncoder instance.
        """

        if self.handle_missing not in ["error", ""]:
            raise ValueError(
                f"Got handle_missing={self.handle_missing}, but expected "
                "any of {'error', ''}. "
            )
        if hasattr(X, "iloc") and X.isna().values.any():
            if self.handle_missing == "error":
                raise ValueError(
                    "Found missing values in input data; set "
                    "handle_missing='' to encode with missing values. "
                )
            else:
                X = X.fillna(self.handle_missing)
        elif not hasattr(X, "dtype") and isinstance(X, list):
            X = np.asarray(X, dtype=object)

        if hasattr(X, "dtype"):
            mask = _object_dtype_isnan(X)
            if X.dtype.kind == "O" and mask.any():
                if self.handle_missing == "error":
                    raise ValueError(
                        "Found missing values in input data; set "
                        "handle_missing='' to encode with missing values. "
                    )
                else:
                    X[mask] = self.handle_missing

        Xlist, n_samples, n_features = self._check_X(X)
        self.n_features_in_ = n_features

        if self.handle_unknown not in ["error", "ignore"]:
            raise ValueError(
                f"Got handle_unknown={self.handle_unknown!r}, but expected "
                "any of {'error', 'ignore'}. "
            )

        if (self.hashing_dim is not None) and (not isinstance(self.hashing_dim, int)):
            raise ValueError(
                f"Got hashing_dim={self.hashing_dim!r}, which has an invalid "
                f"type ({type(self.hashing_dim)}), expected None or int. "
            )

        if self.categories not in ["auto", "most_frequent", "k-means"]:
            for cats in self.categories:
                if not np.all(np.sort(cats) == np.array(cats)):
                    raise ValueError("Unsorted categories are not yet supported. ")

        self.categories_ = list()
        self.random_state_ = check_random_state(self.random_state)

        for i in range(n_features):
            Xi = Xlist[i]
            if self.categories == "auto":
                self.categories_.append(np.unique(Xi))
            elif self.categories == "most_frequent":
                self.categories_.append(self.get_most_frequent(Xi))
            elif self.categories == "k-means":
                uniques, count = np.unique(Xi, return_counts=True)
                self.categories_.append(
                    get_kmeans_prototypes(
                        uniques,
                        self.n_prototypes,
                        sample_weight=count,
                        random_state=self.random_state_,
                    )
                )
            else:
                if self.handle_unknown == "error":
                    valid_mask = np.in1d(Xi, self.categories[i])
                    if not np.all(valid_mask):
                        diff = np.unique(Xi[~valid_mask])
                        raise ValueError(
                            f"Found unknown categories {diff} in column {i} "
                            "during fit. "
                        )
                self.categories_.append(np.array(self.categories[i], dtype=object))

        self.vectorizers_ = []
        self.vocabulary_count_matrices_ = []
        self.vocabulary_ngram_counts_ = []

        for i in range(n_features):
            vectorizer = CountVectorizer(
                analyzer="char",
                ngram_range=self.ngram_range,
                dtype=self.dtype,
                strip_accents=None,
            )

            # Store the raw-categories (and not the preprocessed
            # categories) but use the preprocessed categories to compute
            # the stored count_matrices. This done to preserve the
            # equivalency between the user input and the categories_
            # attribute of the SimilarityEncoder, while being compliant
            # with the CountVectorizer preprocessing steps.
            categories = self.categories_[i]

            self.vectorizers_.append(vectorizer)

            self.vocabulary_count_matrices_.append(
                vectorizer.fit_transform(
                    [preprocess(category) for category in categories]
                )
            )

            self.vocabulary_ngram_counts_.append(
                [
                    get_ngram_count(preprocess(category), self.ngram_range)
                    for category in categories
                ]
            )

        self.drop_idx_ = self._compute_drop_idx()
        if Version(sklearn.__version__) >= Version("1.1.0"):
            self._infrequent_enabled = False

        return self

    def transform(self, X, fast: bool = True) -> np.array:
        """
        Transform X using specified encoding scheme.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to encode.
        fast : bool
            Whether to use the fast computation of ngrams.

        Returns
        -------
        X_new : 2-d array, shape [n_samples, n_features_new]
            Transformed input.
        """

        if hasattr(X, "iloc") and X.isna().values.any():
            if self.handle_missing == "error":
                raise ValueError(
                    "Found missing values in input data; set "
                    "handle_missing='' to encode with missing values. "
                )
            if self.handle_missing != "error":
                X = X.fillna(self.handle_missing)
        elif not hasattr(X, "dtype") and isinstance(X, list):
            X = np.asarray(X, dtype=object)

        if hasattr(X, "dtype"):
            mask = _object_dtype_isnan(X)
            if X.dtype.kind == "O" and mask.any():
                if self.handle_missing == "error":
                    raise ValueError(
                        "Found missing values in input data; set "
                        "handle_missing='' to encode with missing values. "
                    )
                else:
                    X[mask] = self.handle_missing

        Xlist, n_samples, n_features = self._check_X(X)

        for i in range(n_features):
            Xi = Xlist[i]
            valid_mask = np.in1d(Xi, self.categories_[i])

            if not np.all(valid_mask):
                if self.handle_unknown == "error":
                    diff = np.unique(X[~valid_mask, i])
                    raise ValueError(
                        f"Found unknown categories {diff} in column {i} during fit. "
                    )

        min_n, max_n = self.ngram_range

        total_length = sum(len(x) for x in self.categories_)
        out = np.empty((len(X), total_length), dtype=self.dtype)
        last = 0
        for j, categories in enumerate(self.categories_):
            if fast:
                encoded_Xj = self._ngram_similarity_fast(Xlist[j], j)
            else:
                encoded_Xj = ngram_similarity(
                    Xlist[j],
                    categories,
                    ngram_range=(min_n, max_n),
                    hashing_dim=self.hashing_dim,
                    dtype=np.float32,
                )

            out[:, last : last + len(categories)] = encoded_Xj
            last += len(categories)
        return out

    def _ngram_similarity_fast(
        self,
        X: Union[list, np.array],
        col_idx: int,
    ) -> np.array:
        """
        Fast computation of ngram similarity.

        SimilarityEncoder.transform uses the count vectors of the vocabulary in
        its computations. In ngram_similarity, these count vectors have to be
        re-computed each time, which can slow down the execution. In this
        method, the count vectors are recovered from the
        ``vocabulary_count_matrices`` attribute of the SimilarityEncoder,
        speeding up the execution.

        Parameters
        ----------
        X: typing.Union[list, np.array]
            Observations being transformed.
        col_idx: int
            The column index of X in the original feature matrix.
        """
        vectorizer = self.vectorizers_[col_idx]

        unq_X = np.unique(X)
        unq_X_ = np.array([preprocess(x) for x in unq_X])

        X_count_matrix = vectorizer.transform(unq_X_)
        vocabulary_count_matrix = self.vocabulary_count_matrices_[col_idx]
        vocabulary_ngram_count = np.array(
            self.vocabulary_ngram_counts_[col_idx]
        ).reshape(-1, 1)

        se_dict = {}

        Parallel(n_jobs=self.n_jobs, backend="threading")(
            delayed(_ngram_similarity_one_sample_inplace)(
                X_count_vector,
                vocabulary_count_matrix,
                x_str,
                vocabulary_ngram_count,
                se_dict,
                unq_X,
                i,
                self.ngram_range,
            )
            for X_count_vector, x_str, i in zip(
                X_count_matrix, unq_X_, range(len(unq_X))
            )
        )

        out = np.empty(
            (len(X), vocabulary_count_matrix.shape[0]),
            dtype=self.dtype,
        )

        for x, out_row in zip(X, out):
            out_row[:] = se_dict[x]

        return np.nan_to_num(out, copy=False)

    def fit_transform(self, X, y=None, **fit_params) -> np.array:
        """
        Fit SimilarityEncoder to data, then transform it.
        Fits transformer to `X` and `y` with optional parameters
        `fit_params` and returns a transformed version of `X`.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs), \
                default=None
            Target values (None for unsupervised transformations).
        **fit_params
            Additional fit parameters.

        Returns
        -------
        array of shape (n_samples, n_features_new)
            Transformed array.
        """
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.fit(X, y, **fit_params).transform(X)



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

version_file = os.path.join(os.path.dirname(__file__), "dirty_cat", "VERSION.txt")
with open(version_file) as fh:
    VERSION = fh.read().strip()

description_file = os.path.join(os.path.dirname(__file__), "README.rst")
with open(description_file) as fh:
    DESCRIPTION = fh.read()


if __name__ == "__main__":
    setup(
        name="dirty_cat",
        version=VERSION,
        author="Patricio Cerda",
        author_email="patricio.cerda@inria.fr",
        url="http://dirty-cat.github.io/",
        description="Machine learning with dirty categories.",
        long_description=DESCRIPTION,
        license="BSD",
        python_requires=">=3.8",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Topic :: Scientific/Engineering",
            "Topic :: Software Development :: Libraries",
        ],
        platforms="any",
        packages=find_packages(),
        package_data={"dirty_cat": ["VERSION.txt"]},
        install_requires=[
            "scikit-learn>=0.23.0",
            "numpy>=1.17.3",
            "scipy>=1.4.0",
            "pandas>=1.2.0",
            "requests",
            "joblib",
        ],
    )



import numpy as np
import pandas as pd

from dirty_cat.datetime_encoder import DatetimeEncoder


def get_date_array() -> np.array:
    return np.array(
        [
            pd.to_datetime(["2020-01-01", "2020-01-02", "2020-01-03"]),
            pd.to_datetime(["2021-02-03", "2020-02-04", "2021-02-05"]),
            pd.to_datetime(["2022-01-01", "2020-12-25", "2022-01-03"]),
            pd.to_datetime(["2023-02-03", "2020-02-04", "2023-02-05"]),
        ]
    )


def get_constant_date_array() -> np.array:
    return np.array(
        [
            pd.to_datetime(["2020-01-01", "2020-02-04", "2021-02-05"]),
            pd.to_datetime(["2020-01-01", "2020-02-04", "2021-02-05"]),
            pd.to_datetime(["2020-01-01", "2020-02-04", "2021-02-05"]),
            pd.to_datetime(["2020-01-01", "2020-02-04", "2021-02-05"]),
        ]
    )


def get_datetime_array() -> np.array:
    return np.array(
        [
            pd.to_datetime(
                ["2020-01-01 10:12:01", "2020-01-02 10:23:00", "2020-01-03 10:00:00"]
            ),
            pd.to_datetime(
                ["2021-02-03 12:45:23", "2020-02-04 22:12:00", "2021-02-05 12:00:00"]
            ),
            pd.to_datetime(
                ["2022-01-01 23:23:43", "2020-12-25 11:12:00", "2022-01-03 11:00:00"]
            ),
            pd.to_datetime(
                ["2023-02-03 11:12:12", "2020-02-04 08:32:00", "2023-02-05 23:00:00"]
            ),
        ]
    )


def get_dirty_datetime_array() -> np.array:
    return np.array(
        [
            np.array(
                pd.to_datetime(
                    [
                        "2020-01-01 10:12:01",
                        "2020-01-02 10:23:00",
                        "2020-01-03 10:00:00",
                    ]
                )
            ),
            np.array(
                pd.to_datetime([np.nan, "2020-02-04 22:12:00", "2021-02-05 12:00:00"])
            ),
            np.array(
                pd.to_datetime(["2022-01-01 23:23:43", "2020-12-25 11:12:00", pd.NaT])
            ),
            np.array(
                pd.to_datetime(
                    [
                        "2023-02-03 11:12:12",
                        "2020-02-04 08:32:00",
                        "2023-02-05 23:00:00",
                    ]
                )
            ),
        ]
    )


def get_datetime_with_TZ_array() -> pd.DataFrame:
    res = pd.DataFrame(
        [
            pd.to_datetime(["2020-01-01 10:12:01"]),
            pd.to_datetime(["2021-02-03 12:45:23"]),
            pd.to_datetime(["2022-01-01 23:23:43"]),
            pd.to_datetime(["2023-02-03 11:12:12"]),
        ]
    )
    for col in res.columns:
        res[col] = pd.DatetimeIndex(res[col]).tz_localize("Asia/Kolkata")
    return res


def test_fit() -> None:
    # Dates
    X = get_date_array()
    enc = DatetimeEncoder()
    expected_to_extract = ["year", "month", "day", "hour"]
    expected_features_per_column_ = {
        0: ["year", "month", "day"],
        1: ["month", "day"],
        2: ["year", "month", "day"],
    }
    enc.fit(X)
    assert enc._to_extract == expected_to_extract
    assert enc.features_per_column_ == expected_features_per_column_

    X = get_date_array()
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_to_extract = ["year", "month", "day", "hour", "dayofweek"]
    expected_features_per_column_ = {
        0: ["year", "month", "day", "dayofweek"],
        1: ["month", "day", "dayofweek"],
        2: ["year", "month", "day", "dayofweek"],
    }
    enc.fit(X)
    assert enc._to_extract == expected_to_extract
    assert enc.features_per_column_ == expected_features_per_column_

    # Datetimes
    X = get_datetime_array()
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_to_extract = ["year", "month", "day", "hour", "dayofweek"]
    expected_features_per_column_ = {
        0: ["year", "month", "day", "hour", "dayofweek", "total_time"],
        1: ["month", "day", "hour", "dayofweek", "total_time"],
        2: ["year", "month", "day", "hour", "dayofweek"],
    }
    enc.fit(X)
    assert enc._to_extract == expected_to_extract
    assert enc.features_per_column_ == expected_features_per_column_

    X = get_datetime_array()
    enc = DatetimeEncoder(extract_until="minute")
    expected_to_extract = ["year", "month", "day", "hour", "minute"]
    expected_features_per_column_ = {
        0: ["year", "month", "day", "hour", "minute", "total_time"],
        1: ["month", "day", "hour", "minute"],
        2: ["year", "month", "day", "hour"],
    }
    enc.fit(X)
    assert enc._to_extract == expected_to_extract
    assert enc.features_per_column_ == expected_features_per_column_

    # Dirty Datetimes
    X = get_dirty_datetime_array()
    enc = DatetimeEncoder()
    expected_to_extract = ["year", "month", "day", "hour"]
    expected_features_per_column_ = {
        0: ["year", "month", "day", "hour", "total_time"],
        1: ["month", "day", "hour", "total_time"],
        2: ["year", "month", "day", "hour"],
    }
    enc.fit(X)
    assert enc._to_extract == expected_to_extract
    assert enc.features_per_column_ == expected_features_per_column_

    # Datetimes with TZ
    X = get_datetime_with_TZ_array()
    enc = DatetimeEncoder()
    expected_to_extract = ["year", "month", "day", "hour"]
    expected_features_per_column_ = {0: ["year", "month", "day", "hour", "total_time"]}
    enc.fit(X)
    assert enc._to_extract == expected_to_extract
    assert enc.features_per_column_ == expected_features_per_column_

    # Feature names
    # Without column names
    X = get_datetime_array()
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_feature_names = [
        "0_year",
        "0_month",
        "0_day",
        "0_hour",
        "0_dayofweek",
        "0_total_time",
        "1_month",
        "1_day",
        "1_hour",
        "1_dayofweek",
        "1_total_time",
        "2_year",
        "2_month",
        "2_day",
        "2_hour",
        "2_dayofweek",
    ]
    enc.fit(X)
    assert enc.get_feature_names_out() == expected_feature_names

    # With column names
    X = get_datetime_array()
    X = pd.DataFrame(X)
    X.columns = ["col1", "col2", "col3"]
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_feature_names = [
        "col1_year",
        "col1_month",
        "col1_day",
        "col1_hour",
        "col1_dayofweek",
        "col1_total_time",
        "col2_month",
        "col2_day",
        "col2_hour",
        "col2_dayofweek",
        "col2_total_time",
        "col3_year",
        "col3_month",
        "col3_day",
        "col3_hour",
        "col3_dayofweek",
    ]
    enc.fit(X)
    assert enc.get_feature_names_out() == expected_feature_names


def test_transform():
    # Dates
    X = get_date_array()
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_result = np.array(
        [
            [2020, 1, 1, 2, 1, 2, 3, 2020, 1, 3, 4],
            [2021, 2, 3, 2, 2, 4, 1, 2021, 2, 5, 4],
            [2022, 1, 1, 5, 12, 25, 4, 2022, 1, 3, 0],
            [2023, 2, 3, 4, 2, 4, 1, 2023, 2, 5, 6],
        ]
    )
    enc.fit(X)
    assert np.allclose(enc.transform(X), expected_result, equal_nan=True)

    enc = DatetimeEncoder(add_day_of_the_week=False)
    expected_result = np.array(
        [
            [2020, 1, 1, 1, 2, 2020, 1, 3],
            [2021, 2, 3, 2, 4, 2021, 2, 5],
            [2022, 1, 1, 12, 25, 2022, 1, 3],
            [2023, 2, 3, 2, 4, 2023, 2, 5],
        ]
    )
    enc.fit(X)
    assert np.allclose(enc.transform(X), expected_result, equal_nan=True)

    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_result = np.array(
        [
            [2020, 1, 1, 2, 1, 2, 3, 2020, 1, 3, 4],
            [2021, 2, 3, 2, 2, 4, 1, 2021, 2, 5, 4],
            [2022, 1, 1, 5, 12, 25, 4, 2022, 1, 3, 0],
            [2023, 2, 3, 4, 2, 4, 1, 2023, 2, 5, 6],
        ]
    )
    enc.fit(X)
    assert np.allclose(enc.transform(X), expected_result, equal_nan=True)

    # Datetimes
    X = get_datetime_array()[:, 0].reshape(-1, 1)
    enc = DatetimeEncoder(add_day_of_the_week=True)
    # Check that the "total_time" feature is working
    expected_result = np.array(
        [
            [2020, 1, 1, 10, 2, 0],
            [2021, 2, 3, 12, 2, 0],
            [2022, 1, 1, 23, 5, 0],
            [2023, 2, 3, 11, 4, 0],
        ]
    ).astype(np.float64)
    # Time from epochs in seconds
    expected_result[:, 5] = (X.astype("int64") // 1e9).astype(np.float64).reshape(-1)

    enc.fit(X)
    X_trans = enc.transform(X)
    assert np.allclose(X_trans, expected_result, equal_nan=True)

    # Check if we find back the date from the time to epoch
    assert (
        (
            pd.to_datetime(X_trans[:, 5], unit="s") - pd.to_datetime(X.reshape(-1))
        ).total_seconds()
        == 0
    ).all()

    # Dirty datetimes
    X = get_dirty_datetime_array()[:, 0].reshape(-1, 1)
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_result = np.array(
        [
            [2020, 1, 1, 10, 2, 0],
            [np.nan] * 6,
            [2022, 1, 1, 23, 5, 0],
            [2023, 2, 3, 11, 4, 0],
        ]
    )
    # Time from epochs in seconds
    expected_result[:, 5] = (X.astype("int64") // 1e9).astype(np.float64).reshape(-1)
    expected_result[1, 5] = np.nan
    enc.fit(X)
    X_trans = enc.transform(X)
    assert np.allclose(X_trans, expected_result, equal_nan=True)

    # Datetimes with TZ
    # If the dates are timezone-aware, all the feature extractions should
    # be done in the provided timezone.
    # But the full time to epoch should correspond to the true number of
    # seconds between epoch time and the time of the date.
    X = get_datetime_with_TZ_array()
    enc = DatetimeEncoder(add_day_of_the_week=True)
    expected_result = np.array(
        [
            [2020, 1, 1, 10, 2, 0],
            [2021, 2, 3, 12, 2, 0],
            [2022, 1, 1, 23, 5, 0],
            [2023, 2, 3, 11, 4, 0],
        ]
    ).astype(np.float64)
    # Time from epochs in seconds
    expected_result[:, 5] = (
        (X.iloc[:, 0].view(dtype="int64") // 1e9)
        .astype(np.float64)
        .to_numpy()
        .reshape(-1)
    )
    enc.fit(X)
    X_trans = enc.transform(X)
    assert np.allclose(X_trans, expected_result, equal_nan=True)

    # Check if we find back the date from the time to epoch
    assert (
        (
            pd.to_datetime(X_trans[:, 5], unit="s")
            .tz_localize("utc")
            .tz_convert(X.iloc[:, 0][0].tz)
            - pd.DatetimeIndex(X.iloc[:, 0])
        ).total_seconds()
        == 0
    ).all()

    # Check if it's working when the date is constant
    X = get_constant_date_array()
    enc = DatetimeEncoder(add_day_of_the_week=True)
    assert enc.fit_transform(X).shape[1] == 0



import collections
from typing import Any, Hashable, Tuple, Union

import numpy as np
from sklearn.utils import check_array


class LRUDict:
    """dict with limited capacity

    Using LRU eviction avoids memorizing a full dataset"""

    def __init__(self, capacity: int):
        self.capacity = capacity
        self.cache = collections.OrderedDict()

    def __getitem__(self, key: Hashable):
        try:
            value = self.cache.pop(key)
            self.cache[key] = value
            return value
        except KeyError:
            return -1

    def __setitem__(self, key: Hashable, value: Any):
        try:
            self.cache.pop(key)
        except KeyError:
            if len(self.cache) >= self.capacity:
                self.cache.popitem(last=False)
        self.cache[key] = value

    def __contains__(self, key: Hashable):
        return key in self.cache


def check_input(X) -> np.array:
    """
    Check input with sklearn standards.
    Also converts X to a numpy array if not already.
    """
    # TODO check for weird type of input to pass scikit learn tests
    #  without messing with the original type too much

    X_ = check_array(
        X,
        dtype=None,
        ensure_2d=True,
        force_all_finite=False,
    )
    # If the array contains both NaNs and strings, convert to object type
    if X_.dtype.kind in {"U", "S"}:  # contains strings
        if np.any(X_ == "nan"):  # missing value converted to string
            return check_array(
                np.array(X, dtype=object),
                dtype=None,
                ensure_2d=True,
                force_all_finite=False,
            )

    return X_


class Version:
    """
    Replacement for `distutil.version.LooseVersion` and
    `packaging.version.Version`.
    Implemented to avoid `DeprecationWarning`s raised by the former,
    and avoid adding a dependency for the latter.

    It is therefore very bare-bones, so its code shouldn't be too
    hard to understand.
    It currently only supports major and minor versions.

    Inspired from https://stackoverflow.com/a/11887825/9084059
    Should eventually dissapear.

    Examples:
    >>> # Standard usage
    >>> Version(sklearn.__version__) > Version('0.23')
    >>> Version(sklearn.__version__) > '0.23'
    >>> # In general, pass the version as numbers separated by dots.
    >>> Version('1.5') <= Version('1.6.5')
    >>> Version('1.5') <= '1.6.5'
    >>> # You can also pass the separator for specific cases
    >>> Version('1-5', separator='-') == Version('1-6-5', separator='-')
    >>> Version('1-5', separator='-') == '1-6-5'
    >>> Version('1-5', separator='-') == '1.6.5'  # Won't work!
    """

    def __init__(self, value: str, separator: str = "."):
        self.separator = separator
        self.major, self.minor = self._parse_version(value)

    def __repr__(self):
        return f"Version({self.major}.{self.minor})"

    def _parse_version(self, value: str) -> Tuple[int, int]:
        raw_parts = value.split(self.separator)
        if len(raw_parts) == 0:
            raise ValueError(
                f"Could not extract version from {value!r} "
                f"(separator: {self.separator!r})"
            )
        elif len(raw_parts) == 1:
            major = int(raw_parts[0])
            minor = 0
        else:
            major = int(raw_parts[0])
            minor = int(raw_parts[1])
            # Ditch the rest
        return major, minor

    def _cast_to_version(self, other: Union["Version", str]) -> "Version":
        if isinstance(other, str):
            # We pass our separator, as we expect they are the same
            other = Version(other, self.separator)
        return other

    def __eq__(self, other: Union["Version", str]):
        other = self._cast_to_version(other)
        return (self.major, self.minor) == (other.major, other.minor)

    def __ne__(self, other: Union["Version", str]):
        other = self._cast_to_version(other)
        return (self.major, self.minor) != (other.major, other.minor)

    def __lt__(self, other: Union["Version", str]):
        other = self._cast_to_version(other)
        return (self.major, self.minor) < (other.major, other.minor)

    def __le__(self, other: Union["Version", str]):
        other = self._cast_to_version(other)
        return (self.major, self.minor) <= (other.major, other.minor)

    def __gt__(self, other: Union["Version", str]):
        other = self._cast_to_version(other)
        return (self.major, self.minor) > (other.major, other.minor)

    def __ge__(self, other: Union["Version", str]):
        other = self._cast_to_version(other)
        return (self.major, self.minor) >= (other.major, other.minor)
